/***********************************************************************
 * File: days.c
 * 	(Extracted from
 *	   "$Id: days.c,v 1.10 2006/05/05 08:25:08 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements "is_valid_date()" and "date_in_English()"
 * API-s.
 *
 * Copyright (C) 2005 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "days.h"

/* get the year, month and day from a numeric string. */
static int get_ymd(char		*date
		, size_t	sz
		, unsigned long	*year
		, int		*month
		, int		*day
		)
{
	char	sv, *endp = NULL;

	/* test: the arguments are illegal? */
	if ((((((date == NULL)
	|| sz < 5)
	|| !isdigit(date[sz - 1]))
	|| !isdigit(date[sz - 2]))
	|| !isdigit(date[sz - 3]))
	|| !isdigit(date[sz - 4])) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* set the month and day pointing areas */
	*month	= (10 * (date[sz - 4] - '0'))
				+ (date[sz - 3] - '0');
	*day	= (10 * (date[sz - 2] - '0'))
				+ (date[sz - 1] - '0');

	/* test: the month is illegal? */
	if ((((*month < 1) || *month > 12)
	|| *day < 1) || *day > 31) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* convert the year part of the date string
	 * into the unsigned long
	 */
	sv	= date[sz - 4];
	date[sz - 4]	= '\0';
	errno	= 0; /* initialize errno */
	*year	= strtoul((const char *)date, &endp, 10);
	date[sz - 4]	= sv;

	/* test: the year part is over flow? */
	if ((*year == ULONG_MAX) && errno == ERANGE) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* test: the size of date string is least than 8 (yyyymmdd)? */
	if (sz < 8) {
		/* amend as the 21st century */
		*year	+= DAYS_BASE_YEAR;
	}

	return	(*year < 1)? DAYS_ILLEGAL_ARGS: 0;
}


/* This determines the numeric string means a valid date. */
static int _is_valid_date(char *date
			, size_t sz
			, unsigned long *year
			, int *month
			, int *day
			)
{
	int	result, days[] = { 31, 28, 31, 30, 31, 30
				 , 31, 31, 30, 31, 30, 31, 29
				 };

	/* get the year, month and day from date string */
	if (get_ymd(date, sz, year, month, day) != 0) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* test: the tuple of year, month and day is valid? */
	if (*month == 1 || *month > 2) {
		result	= (*day <= days[*month - 1])?
					DAYS_TRUE: DAYS_FALSE;
	}
	else if (IS_LEAP_YEAR(*year)) {
		result	= (*day <= days[12])? DAYS_TRUE: DAYS_FALSE;
	}
	else {
		result	= (*day <= days[1]) ? DAYS_TRUE: DAYS_FALSE;
	}

	return	result; /* DAYS_TRUE:  valid date string
			 * DAYS_FALSE: invalid date string
			 */
}


/*
 * This API determines the numeric string is valid date (yyy...yyymmdd)
 *
 * arguments:
 *	date is pointing to the date string (yyy...yyymmdd).
 *	sz indicates the size of str pointing area.
 *
 * returns:
 *	DAYS_ILLEGAL_ARGS (= -1) : illegal arguments
 *	DAYS_FALSE (= 0)         : the string doesn't means valid date.
 *	DAYS_TRUE  (= 1)         : the string means valid date.
 */
int is_valid_date(char *date, size_t sz)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;

	return	_is_valid_date(date, sz, &year, &month, &day);
}


/*
 * This API translate the numeric string into date in English.
 *
 * arguments:
 *	date is pointing to the numeric string (yyy...yyymmdd).
 *	sz indicates the size of str pointing area.
 *	fmt indicates the format of European or American.
 *	str is pointing the output string.
 *	str_sz indicates the size of the str output area.
 *
 * returns:
 *	NULL : illegal arguments
 *	non-NULL : the translated string address (same as str)
 */
char *date_in_English(char *date
		, size_t sz
		, DAYS_FMT fmt
		, char *str
		, size_t str_sz
		)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;
	char		*mon_names[] = {
				  "January", "February", "March"
				, "April",   "May",      "June"
				, "July",    "August",   "September"
				, "October", "November", "December"
				};

	/* test: arguments are illegal? */
	if (((((str == NULL)
	|| fmt < DAYS_FMT_EUROPEAN)
	|| fmt > DAYS_FMT_AMERICAN)
	|| _is_valid_date(date, sz, &year, &month, &day) != DAYS_TRUE)
	|| str_sz < ((sz < 8)? 8: sz) + strlen(mon_names[month - 1])
					+ ((day > 10)? 2: 1)) {
		return	NULL;
	}

	/* which format? */
	if (fmt == DAYS_FMT_EUROPEAN) {
		snprintf(str, str_sz
			, ((day > 10)? "%2d, %s %lu": "%1d, %s, %lu")
			, day, mon_names[month - 1], year
			);
	}
	else {	/* in American English */
		snprintf(str, str_sz
			, ((day > 10)? "%s, %2d %lu": "%s,%2d, %lu")
			, mon_names[month - 1], day, year
			);
	}

	return	str; /* translated date string */
}


/* table of the past days of the month */
static int	past_days[] =		{   0,  31,  59,  90
					, 120, 151, 181, 212
					, 243, 273, 304, 334
					, 365
					}
		, leap_past_days[] =	{   0,  31,  60,  91
					, 121, 152, 182, 213
					, 244, 274, 305, 335
					, 366
					};

/* This returns the day of year for year/month/day. */
static int _get_day_of_year(unsigned long year, int month, int day)
{
	return	past_days[month - 1]
		+ day
		+ (((month > 2) && IS_LEAP_YEAR(year))? 1LL: 0LL)
		;
}


/*
 * This API returns the day of year.
 *
 * arguments:
 *	date is pointing to the date string.
 *	sz indicates the length of the date string.
 *
 * returns:
 *	zero : illegal arguments
 *	positive integers: the Julian date for date string
 */
int get_day_of_year(char *date, size_t sz)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;

	/* test: arguments are illegal? */
	if (((date == NULL)
	|| sz == 0)
	|| _is_valid_date(date, sz, &year, &month, &day) != DAYS_TRUE) {
		return	0;
	}

	return	_get_day_of_year(year, month, day);
}

/*
 * This API counts the days since 1, January AD.1 (it's the 1st day).
 *
 * arguments:
 *	date is pointing to the date.
 *	sz indicates the length of the date string.
 *
 * returns:
 *	LLONG_MIN :		illegal arguments
 *	0LL to 1947821355738LL:	the count of days.
 */
long long days_count(char *date, size_t sz)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;
	long long	result;

	if ((date == NULL)
	|| _is_valid_date(date, sz, &year, &month, &day) != DAYS_TRUE) {
		return	LLONG_MIN; /* Failed: illegal arguments */
	}

	result	= (long long)(year - 1);
	result	= result * 365LL + result / 400LL
				- result / 100LL + result / 4LL;
	return	result	+ _get_day_of_year(year, month, day);
}


/*
 * This API subtracts the day-counts of date1 and date2.
 *
 * arguments:
 *	date1 is pointing to the subtractive date string.
 *	sz1 indicates the length of date1 staring.
 *	date2 is pointing to the subtracter date string.
 *	sz2 indicates the length of date2 staring.
 *
 * returns:
 *	LLONG_MIN : illegal arguments
 *	-1947821355738LL to 1947821355738LL: the diff of day-counts.
 */
long long diff_dates(char *date1, size_t sz1, char *date2, size_t sz2)
{
	long long	d1, d2;

	if (((d1 = days_count(date1, sz1)) == LLONG_MIN)
	|| (d2 = days_count(date2, sz2)) == LLONG_MIN) {
		return	LLONG_MIN;
	}

	return	d1 - d2;
}


/*
 * This API translate day-count to the 'YYY...YYYMMDD' formatted string.
 *
 * arguments:
 *	  day_count indicates the count of past days that returns of
 *	`days_count()' API.
 *	  date points to the buffer that is outputted 'YYY...YYYMMDD'
 *	formatted string.
 *	  sz indicates the size of buffer that is pointed from `date`.
 *	sz must be greater than nine bytes.
 *
 * returns:
 *	NULL: invalid arguments
 *	non-NULL: normal end (same as `date`)
 */
char *days_count_to_date_str(long long day_count, char *date, size_t sz)
{
	int		*past;
	size_t		m = 0;
	long long	year = day_count / 365
			, d = day_count % 365
			;

	/* test: the arguments are illegal? */
	if (((day_count <= 0LL) || date == NULL) || sz < 9) {
		errno	= EINVAL;
		return NULL; /* ERROR: invalid arguments */
	}

	/* decrement `d` for leap years */
	d	-= ((year / 4) - (year / 100) + (year / 400));

	/* adjust `d` and `year` for a long long further year... */
	while (d <= 0LL) {
		/* decrement `d` for days of `year` */
		d	+= ((IS_LEAP_YEAR(year))? 366: 365);
		--year; /* decrement `year` */
	}

	/* now, `year` indicates the last year */
	++year; /* increment `year` */

	/* determine month `m`... */
	for (past = (IS_LEAP_YEAR(year))? leap_past_days: past_days
		, m = 1
	; (m <= 12) && d > past[m]
	; ++m) { ; }

	d	-= past[m - 1]; /* determine day `d` */

	/* if year indicates between the 1st and 10th centuries, we must
	 * format it to the zero filled 4 digits.  otherwise, we must
	 * format it as is.
	 */
	snprintf(date, sz
		, ((year < 1000)? "%04lld%02d%02lld"
				: "%lld%02d%02lld")
		,			year, m, d
		);


	return	date; /* normal end (address of output buffer `date`) */
}
