/***********************************************************************
 * File: numeral.c
 * 	(Extracted from
 *	   "$Id: numeral.c,v 1.7 2005/11/04 15:17:12 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements `numeral()' API.
 *
 * Copyright (C) 2005 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "numeral.h"

/* try to put the name of triad digits */
static int try_to_put_numeral(char *naming
				, char *delimit
				, size_t *sz
				, char **dst
				)
{
	size_t	len;

	if (strlen(delimit) + strlen(naming) > *sz - 1) {
		/* buffer overflow */
		return	NUMERAL_BUFFER_OVERFLOW;
	}
	snprintf(*dst, *sz, "%s%s", delimit, naming);
	*sz	-= (len = strlen(*dst));
	*dst	+= len;
	strcpy(delimit, " ");

	return	NUMERAL_NORMAL_END;
}

/* translate numerals (from zero to 999) */
static int triad(char *src
	, size_t digit
	, char *dst
	, size_t sz
	, int *n
	, char *delimit
	)
{
	static char *hundred_namings[] = {
			  "zero", "one", "two", "three", "four"
			, "five", "six", "seven", "eight", "nine"
			};
	static char *ten_namings[] = {
			  "", "", "twenty", "thirty", "forty", "fifty"
			, "sixty", "seventy", "eighty", "ninety"
			};
	static char *teen_namings[] = {
			  "ten", "eleven", "twelve", "thirteen"
			, "fourteen", "fifteen", "sixteen", "seventeen"
			, "eighteen", "nineteen"
			};
	size_t	i, len, t;
	char	*naming;
	int	result;

	/* get the number */
	for (*n = 0, i = 0; i < digit; ++i) {
		if (!isdigit(src[i])) {
			/* illegal arguments */
			return	NUMERAL_ILLEGAL_ARGS;
		}
		*n	= 10 * *n + src[i] - '0';
	}

	/* test: the number is zero? */
	if (*n == 0) {
		/* test: the first digits describes "zero"? */
		if (*delimit == '\0') {
			/* try to put the "zero" to `dst' string */
			naming	= hundred_namings[0];
			if (strlen(naming) > sz - 1) {
				/* buffer overflow */
				return	NUMERAL_BUFFER_OVERFLOW;
			}
			snprintf(dst, sz, "%s", naming);
		}
		return	NUMERAL_NORMAL_END; /* normal end (zero) */
	}

	/* test: the number is greater than 99? */
	if (*n / 100 > 0) {
		/* try to put "hundred(s)" */
		t	= *n / 100;
		naming	= (t == 1)? "hundred": "hundreds";
		if (strlen(delimit) + strlen(hundred_namings[t])
					+ strlen(naming) + 1 > sz - 1) {
			/* buffer overflow */
			return	NUMERAL_BUFFER_OVERFLOW;
		}
		snprintf(dst, sz, "%s%s %s"
			,	delimit, hundred_namings[t], naming
			);
		sz	-= (len = strlen(dst));
		dst	+= len;
		strcpy(delimit, " ");
	}

	/* test: the number is between 20 and 99? */
	if (*n % 100 > 19) {
		t	= (*n % 100) / 10;
		if ((result = try_to_put_numeral(ten_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
			return	result; /* buffer over flow */
		}

		/* test: the rest of number exist? */
		if (*n % 10 > 0) {
			/* the number is between 1 and 9 */
			t	= *n % 10;
			if ((result = try_to_put_numeral(
					  hundred_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
				return	result; /* buffer over flow */
			}
		}
	}
	else if (*n % 100 > 9) {
		/* the number is between 10 and 19 */
		t	= *n % 10;
		if ((result = try_to_put_numeral(teen_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
			return	result; /* buffer over flow */
		}
	}
	else if (*n % 10 > 0) {
		/* the number is between 1 and 9 */
		t	= *n % 10;
		if ((result = try_to_put_numeral(hundred_namings[t]
					, delimit
					, &sz
					, &dst
					)) != NUMERAL_NORMAL_END) {
			return	result; /* buffer over flow */
		}
	}

	return	NUMERAL_NORMAL_END;
}


/* translate a numeral string to the English or American numeral
 * representation
 */
struct numeral_naming_t {
		size_t	len[2];		/* as the exponent */
		char	*names[2];	/* names of the digits */
};

static int translate(struct numeral_naming_t *defs
			, size_t defs_n
			, char **src
			, size_t *srclen
			, char **dst
			, size_t *sz
			, int *n
			, char *delimit
			, NUMERAL_REPRE_STYLE_t style
			)
{
	size_t	i, dstlen, src_len;
	int	m, result;
	char	*src_sv, *naming;

	/* try to all namings those are defined in `defs[]'... */
	for (i = 0, m = 0; i < defs_n; ++i) {
		/* initialize src_sz and n */
		src_sv	= *src;
		*n	= 0;

		/* test: the length of src is greater than this name? */
		if (*srclen > defs[i].len[style]) {
			/* get triad digits length */
			src_len = *srclen - defs[i].len[style];

			/* translate the three digits */
			if ((style == NUMERAL_STYLE_ENGLISH)
				&& defs[i].len[style] >= 6) {
				m	= 0; /* initialize m */
				result = translate(defs
						, defs_n
						, &src_sv
						, &src_len
						, dst
						, sz
						, &m
						, delimit
						, style
						);
			}
			else {
				result = triad(*src
						, src_len
						, *dst
						, *sz
						, n
						, delimit
						);
				if ((style == NUMERAL_STYLE_ENGLISH)
				&& defs[i].len[style] == 3) {
					m	= *n;
				}
			}
			
			if (result != NUMERAL_NORMAL_END) {
				return	result; /* Failed */
			}

			/* test: the three digits is greater than zero?
			 */
			if ((m > 0) || *n > 0) {
				/* try to put the name */
				*sz	-= (dstlen = strlen(*dst));
				*dst	+= dstlen;
				naming	=
					defs[i].names[(m > 1 || *n > 1)?
								0: 1
							];
				if (strlen(delimit) + strlen(naming)
							> *sz - 1) {
					/* Failed: buffer overflow */
					return	NUMERAL_BUFFER_OVERFLOW;
				}
				if (defs[i].len[style] > 0) {
					snprintf(*dst, *sz, "%s%s"
						,	delimit, naming
						);
					*sz -= (dstlen = strlen(*dst));
					*dst += dstlen;
				}
			}

			/* increment the src pointer */
			*src	+= (*srclen - defs[i].len[style]);
			*srclen	= defs[i].len[style];
		}
	}

	/* output the value of three or six digits */
	*n	= m * 1000 + *n;

	return	NUMERAL_NORMAL_END;
}

/*
 * This API translates a numeral string to English or American notation.
 *
 * arguments:
 *	src is pointing to the numeral string.
 *	dst is pointing to output string area.
 *	sz indicates the size of `dst' pointing to area.
 *	style indicates the numeral representation style for English
 *	or American.
 *
 * returns:
 *	NUMERAL_NORMAL_END (= 0)	: normal end
 *	NUMERAL_ILLEGAL_ARGS (= 1)	: illegal arguments
 *	NUMERAL_BUFFER_OVERFLOW (= 2)	: buffer overflow
 */
int numeral(char *src
	, char *dst
	, size_t sz
	, NUMERAL_REPRE_STYLE_t style
	)
{
	static struct numeral_naming_t defs[] = {
		  { {120, 63}
		  	, { "vigintillions",	  "vigintillion" } }
		, { {114, 60}
			, { "novemdecillions",	  "novemdecillion" } }
		, { {108, 57}
			, { "octodecillions",	  "octodecillion" } }
		, { {102, 54}
			, { "septendecillions",	  "septendecillion" } }
		, { {96, 51}
			, { "sexdecillions",	  "sexdecillion" } }
		, { {90, 48}
			, { "quindecillions",	  "quindecillion" } }
		, { {84, 45}
			, { "quattuordecillions", "quattuordecillion" }}
		, { {78, 42}
			, { "tredecillions",	  "tredecillion" } }
		, { {72, 39}
			, { "duodecillions",	  "duodecillion" } }
		, { {66, 36}
			, { "undecillions",	  "undecillion" } }
		, { {60, 33}
			, { "decillions",	  "decillion" } }
		, { {54, 30}
			, { "nonillions",	  "nonillion" } }
		, { {48, 27}
			, { "octillions",	  "octillion" } }
		, { {42, 24}
			, { "septillions",	  "septillion" } }
		, { {36, 21}
			, { "sextillions",	  "sextillion" } }
		, { {30, 18}
			, { "quintillions",	  "quintillion" } }
		, { {24, 15}
			, { "quadrillions",	  "quadrillion" } }
		, { {18, 12}
			, { "trillions",	  "trillion" } }
		, { {12, 9}
			,  { "billions",	  "billion" } }
		, { {6, 6}
			,  { "millions",	  "million" } }
		, { {3, 3}
			,  { "thousands",	  "thousand" } }
		, { {0, 0}
			,  { "",         	  "" } }
		};
	size_t	srclen, dstlen;
	char	delimit[2], *naming;
	int	n = 0, result;

	/* test: illegal arguments? */
	if (((src == NULL)
	|| dst == NULL)
	|| sz < 2) {
		/* Failed: illegal arguments */
		return	NUMERAL_ILLEGAL_ARGS;
	}

	/* get length of `src' string */
	srclen = strlen(src);

	/* test: the first column is a sign? */
	if (*src == '+') {
		/* skip the positive sign */
		++src, --srclen;
	}
	else if (*src == '-') {
		/* try to put "minus " */
		naming	= "minus ";
		if (strlen(naming) > sz - 1) {
			/* Failed: buffer overflow */
			return	NUMERAL_BUFFER_OVERFLOW;
		}
		snprintf(dst, sz, "%s",	naming);
		sz -= (dstlen = strlen(dst));
		dst += dstlen;

		/* skip the negative sign */
		++src, --srclen;
	}

	/* test: the length of src is illegal? */
	if ((srclen == 0)
	|| srclen > defs[0].len[style]
			+ ((style == NUMERAL_STYLE_ENGLISH)? 6: 3)) {
		/* Failed: illegal arguments */
		return	NUMERAL_ILLEGAL_ARGS;
	}

	/* initialize the delimit */
	*delimit	= '\0';

	/* translate the digits to English or American numeral
	 * representation
	 */
	result	= translate(defs
			, sizeof(defs) / sizeof(defs[0])
			, &src
			, &srclen
			, &dst
			, &sz
			, &n
			, delimit
			, style
			);

	return	NUMERAL_NORMAL_END; /* translated the number */
}
